import "Turbine";
import "Turbine.Gameplay";
import "MinstrelBuffs.MinstrelBuffWindow.Strings";
import "MinstrelBuffs.MinstrelBuffWindow.BuffWindow";
import "MinstrelBuffs.MinstrelBuffWindow.MinstrelEffects";
import "MinstrelBuffs.MinstrelBuffWindow.MinstrelBuffCommand";
import "MinstrelBuffs.MinstrelBuffWindow.VindarPatch"
import "MinstrelBuffs.MinstrelBuffWindow.Settings";
import "MinstrelBuffs.MinstrelBuffWindow.EventHelper";
import "MinstrelBuffs.MinstrelBuffWindow.OptionWindow";
import "MinstrelBuffs.MinstrelBuffWindow.SoliloquyTrackerRow";
import "MinstrelBuffs.MinstrelBuffWindow.SoliloquyTrackerWindow";
import "MinstrelBuffs.MinstrelBuffWindow.MelodicInterludeWindow";

LocalPlayer = Turbine.Gameplay.LocalPlayer.GetInstance();
EffectList = LocalPlayer:GetEffects();

MinstrelBuffMain = class();

DEBUG_IGNORE_BALLAD_EFFECTS = false;

KEY_ACTION_TOGGLE_HUD = 0x100000B3;    -- F12
KEY_ACTION_REPOSITION_UI = 0x1000007B; -- Ctrl + \

-- Creates the main class.
function MinstrelBuffMain:Constructor()
	self.settings = MinstrelBuffs.MinstrelBuffWindow.Settings();
	self:RefreshMounted() -- Maybe Minstrelbuff is loaded when mounted.
	self.BuffWindow = MinstrelBuffs.MinstrelBuffWindow.BuffWindow(self.settings);
	self.SoliloquyWindow = MinstrelBuffs.MinstrelBuffWindow.SoliloquyTrackerWindow(self.settings);
	self.MelodicInterludeWindow = MinstrelBuffs.MinstrelBuffWindow.MelodicInterludeWindow(self.settings);
	self.optionWindow = OptionWindow(self.settings);
	self.balladsActive = false;
	self.seriousBusinessEffect = _LANG.EFFECTS.SERIOUS_BUSINESS;
	self.checkForSeriousBusiness = self.settings:GetSetting("CheckForSeriousBusiness");
	self.BuffWindow:SetCheckForSeriousBusiness(self.checkForSeriousBusiness);
	self.showWarSpeechTimers = self.settings:GetSetting("ShowWarSpeechTimers");
	self.showMelodicInterlude = self.settings:GetSetting("ShowMelodicInterlude");

	-- Options support for Turbine PluginManager
	plugin.GetOptionsPanel = function() return self:GetOptionControl(); end

	self.settings:AddCallback(
		"CheckForSeriousBusiness",
		function()
			self.checkForSeriousBusiness = self.settings:GetSetting("CheckForSeriousBusiness");
			self.BuffWindow:SetCheckForSeriousBusiness(self.checkForSeriousBusiness);
		end	);

	self.settings:AddCallback(
		"ShowWarSpeechTimers",
		function()
			self.showWarSpeechTimers = self.settings:GetSetting("ShowWarSpeechTimers");
		end	);

	self.settings:AddCallback(
		"ShowMelodicInterlude",
		function()
			self.showMelodicInterlude = self.settings:GetSetting("ShowMelodicInterlude");
		end	);

end

-- Writes all current buffs to the chat window.
function MinstrelBuffMain:WriteBuffsToChat()
	local placeCount = self:GetBuffCount();
	local effectCount = EffectList:GetCount();
	Turbine.Shell.WriteLine( "Current buffs: " .. placeCount );
	for	i = 1, effectCount, 1 do
		local effect = EffectList:Get(i);
		Turbine.Shell.WriteLine( effect:GetName() );	
	--[[
		if (i == 1) then
			effect.test = {};
			for k,v in pairs(effect) do
				Turbine.Shell.WriteLine( k .. " - " .. type(v));
				if (k == "__index" or k == "__metatable") then
					for k2,v2 in pairs(v) do
						Turbine.Shell.WriteLine( "   " .. k2 .. " - " .. type(v2));
						if (type(v2) == "table") then
							for k3,v3 in pairs(v2) do
								Turbine.Shell.WriteLine( "      " .. k3 .. " - " .. type(v3));
							end
						end
					end
				end
			end
		end
	end

	Turbine.Shell.WriteLine( "members" );
	for k,v in pairs(self) do
		Turbine.Shell.WriteLine( k .. "" );
	--]]
	end
end

function MinstrelBuffMain:HandleCombatChanged(isInCombat)
	self:RefreshVisibility();
	self.SoliloquyWindow:HandleCombatChanged(isInCombat);
end

-- Prepares all event handling and shows the buff window.
function MinstrelBuffMain:ShowWindow()
	self.EffectAddedCallback = function(sender, args) self:EffectAdded(sender, args); end
	self.EffectRemovedCallback = function(sender, args) self:EffectRemoved(sender, args); end
	self.EffectsClearedCallback = function(sender, args) self:EffectsCleared(sender, args); end
	self.InCombatChanged = function(sender, args)
		local isInCombat = LocalPlayer:IsInCombat();
		self:HandleCombatChanged(isInCombat);
	end
	self.MountChanged = function(sender, args) self:RefreshMounted(); end
	self.MinstrelStanceChangedCallback = function(sender, args) self:MinstrelStanceChanged(); end
	self.CodaSkillResetTimeChangedCallback = function(sender) self:CodaSkillResetTimeChanged(); end
	self.CrySkillResetTimeChangedCallback = function(sender) self:CrySkillResetTimeChanged(); end
	self.MajorBalladResetTimeChangedCallback = function(sender) self:MajorBalladResetTimeChanged(); end
	self.codaWasPlayed = false;
	self.codaSkillList = { };
	self.crySkillList = { };
	self.majorBalladSkillList = { };
	
	-- Add buff, combat and warsteed events.
	AddCallback(EffectList, "EffectAdded", self.EffectAddedCallback);
	AddCallback(EffectList, "EffectRemoved", self.EffectRemovedCallback);
	AddCallback(EffectList, "EffectCleared", self.EffectsClearedCallback);
	AddCallback(LocalPlayer, "InCombatChanged", self.InCombatChanged);
	AddCallback(LocalPlayer, "MountChanged", self.MountChanged);
	if (LocalPlayer:GetClass() == Turbine.Gameplay.Class.Minstrel) then
		-- Add stance events
		self.classAttributes = LocalPlayer:GetClassAttributes();
		AddCallback(self.classAttributes, "StanceChanged", self.MinstrelStanceChangedCallback);
		
		-- Add coda events.
		local skillList = LocalPlayer:GetTrainedSkills();
		local skillCount = skillList:GetCount();
		--Turbine.Shell.WriteLine( "Current trained skills: " .. skillCount );
		local codaSkillListIndex = 1;
		local crySkillListIndex = 1;
		local majorBalladSkillListIndex = 1;
		for	i = 1, skillCount, 1 do
			local skill = skillList:GetItem(i);
			if (self:IsValidCodaSkill(skill)) then
				self.codaSkillList[codaSkillListIndex] = skill;
				codaSkillListIndex = codaSkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.CodaSkillResetTimeChangedCallback);
			end

			if (self:IsValidCrySkill(skill)) then
				self.crySkillList[crySkillListIndex] = skill;
				crySkillListIndex = crySkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.CrySkillResetTimeChangedCallback);
			end

			if (self:IsValidMajorBalladSkill(skill)) then
				self.majorBalladSkillList[majorBalladSkillListIndex] = skill;
				majorBalladSkillListIndex = majorBalladSkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.MajorBalladResetTimeChangedCallback);
			end
		end		
	end

	-- Add event for position changing.
	self.PositionChangedCallback = function(sender, args) self:SaveMainPositionSettings(); end
	AddCallback(self.BuffWindow, "PositionChanged", self.PositionChangedCallback);

	-- Show the window.
	self.BuffWindow:Start();
	self:RefreshEffectDisplay();
end

-- Removes all events and hides the buff window.
function MinstrelBuffMain:HideWindow()
	if (self.EffectAddedCallback ~= nil) then
		RemoveCallback(EffectList, "EffectAdded", self.EffectAddedCallback);
		RemoveCallback(EffectList, "EffectRemoved", self.EffectRemovedCallback);
		RemoveCallback(EffectList, "EffectCleared", self.EffectsClearedCallback);
		RemoveCallback(LocalPlayer, "InCombatChanged", self.InCombatChanged);
		RemoveCallback(LocalPlayer, "MountChanged", self.MountChanged);
		if (self.classAttributes ~= nil) then
			RemoveCallback(self.classAttributes, "StanceChanged", self.MinstrelStanceChangedCallback);
			self.classAttributes = nil;
		end
		for i = 1, #self.codaSkillList, 1 do
			RemoveCallback(self.codaSkillList[i], "ResetTimeChanged", self.CodaSkillResetTimeChangedCallback);
		end
		for i = 1, #self.crySkillList, 1 do
			RemoveCallback(self.crySkillList[i], "ResetTimeChanged", self.CrySkillResetTimeChangedCallback);
		end
		for i = 1, #self.majorBalladSkillList, 1 do
			RemoveCallback(self.majorBalladSkillList[i], "ResetTimeChanged", self.MajorBalladResetTimeChanged);
		end
		self.EffectAddedCallback = nil;
		self.EffectRemovedCallback = nil;
		self.EffectsClearedCallback = nil;
	end	
	
	if (self.PositionChangedCallback ~= nil) then
		RemoveCallback(self.BuffWindow, "PositionChanged", self.PositionChangedCallback);
		self.PositionChangedCallback = nil;
	end
	self.BuffWindow:Stop();
end

function MinstrelBuffMain:GetEffectFromEffectCallback(args)
	local effect = nil;
	if (args.Effect ~= nil) then
		effect = args.Effect;
	else
		if (EffectList:GetCount() >= args.Index) then
			effect = EffectList:Get(args.Index);
		end
	end
	return effect;
end

function MinstrelBuffMain:EffectAdded(sender, args)
	local effect = self:GetEffectFromEffectCallback(args);

	if (effect ~= nil) then
		local name = effect:GetName();
		if (self.showWarSpeechTimers and self:IsValidWarSpeechEffect(name)) then -- need a check for warrior-skald
			self.BuffWindow:WarSpeechAdded();
		elseif(self.showMelodicInterlude and self:IsValidMelodicInterludeEffect(name)) then
			self.MelodicInterludeWindow:MelodicInterludeStart();
		end
	end

	self:RefreshEffectDisplay();
end

function MinstrelBuffMain:EffectRemoved(sender, args)
	local effect = self:GetEffectFromEffectCallback(args);

	if (effect ~= nil) then
		local name = effect:GetName();
		if (self:IsValidWarSpeechEffect(name)) then
			self.BuffWindow:WarSpeechRemoved();
		elseif(self:IsValidMelodicInterludeEffect(name)) then
			self.MelodicInterludeWindow:MelodicInterludeStop();
		end
	end

	self:RefreshEffectDisplay();
end

function MinstrelBuffMain:EffectsCleared(sender, args)
	self:RefreshEffectDisplay();
end

-- Locks or unlocks the ui for positioning (CTRL + #)
-- locked - boolean
function MinstrelBuffMain:UiLocked(locked)
	self:SaveMainPositionSettings();
	self.BuffWindow:UiLocked(locked);
end

-- Opens the options window (in the turbine plugin manager)
function MinstrelBuffMain:ShowOptions()
	Turbine.PluginManager.ShowOptions(Plugins.MinstrelBuff);
end

-- Returns the options control. Used by turbines plugin manager.
function MinstrelBuffMain:GetOptionControl()
	return self.optionWindow:GetOptionControl();
end

-- Saves the position of the window.
function MinstrelBuffMain:SaveMainPositionSettings()
	--Turbine.Shell.WriteLine("SavePosition");
	if (self.BuffWindow ~= nil) then
		self.BuffWindow:SavePosition();
	end
end

-- Returns the localized effect names.
function MinstrelBuffMain:GetLocalization()
	return self.settings:GetEffectLocalizationNames();
end

-- Occurs, when a buff effect is added or removed.
-- Refreshes the BuffWindow contents.
function MinstrelBuffMain:RefreshEffectDisplay()
	local effectCount = EffectList:GetCount();
	local balladEffectPlace = 1;
	local anthemEffectPlace = 1;
	local sortedEffects = {};
	local isAnthemFixed = self.settings:GetFixedAnthemPositions();
	local showAnthems = self.settings:GetShowAnthems();
	local isSeriousBusiness = false;
	for	i = 1, effectCount, 1 do
		sortedEffects[i] = EffectList:Get(i);

		if (self.checkForSeriousBusiness and
			not isSeriousBusiness and
			sortedEffects[i]:GetName() == self.seriousBusinessEffect) then
			isSeriousBusiness = true;
		end
	end
	table.sort(sortedEffects, function(a,b) return a:GetStartTime() < b:GetStartTime() end);

	if (self.checkForSeriousBusiness) then
		self.BuffWindow:SetSeriousBusiness(isSeriousBusiness);
	end

	if (isAnthemFixed == true) then
		for i = 1, 6, 1 do
			self.BuffWindow:HideAnthem(i);
		end
	end

	for	i = 1, effectCount, 1 do
		--local effect = EffectList:Get(i);
		local effect = sortedEffects[i];
		local effectName = effect:GetName();
		if (balladEffectPlace <= 3 and self:IsValidBalladEffect(effectName)
			and not DEBUG_IGNORE_BALLAD_EFFECTS) then
			self.BuffWindow:ShowEffect(effect, balladEffectPlace);
			balladEffectPlace = balladEffectPlace + 1;
		end
		if (showAnthems == true) then
			if (isAnthemFixed == false) then
				if (anthemEffectPlace <= 6 and self:IsValidAnthemEffect(effectName)) then
					-- Maybe the BuffWindow could control where the anthems go?
					self.BuffWindow:ShowAnthem(effect, anthemEffectPlace);
					anthemEffectPlace = anthemEffectPlace + 1;
				end
			else
				if (self:IsValidAnthemEffect(effectName)) then
					local anthemEffectPosition = self:GetAnthemEffectPosition(effectName);
					self.BuffWindow:ShowAnthem(effect, anthemEffectPosition);
					-- Effect border only works with fixed anthems.
					self.BuffWindow:ShowAnthemBorder(anthemEffectPosition);
				end
			end
		end
	end

	if (balladEffectPlace == 1 and self.balladsActive) then
		-- Went from having ballads to not having ballads
		self.balladsActive = false;
		-- Make sure the combat timer stopped
		self.BuffWindow:BalladEnded();
	elseif (balladEffectPlace > 1 and not self.balladsActive) then
		-- Went from not having ballads to having ballads
		self.balladsActive = true;
	end
	
	if (balladEffectPlace <= 3) then
		for i = balladEffectPlace, 3, 1 do
			self.BuffWindow:HideEffect(i);
		end
		if (balladEffectPlace == 1 and self.codaWasPlayed == true) then -- Coda was played, hide all borders.
			self.codaWasPlayed = false;
			self.BuffWindow:HideAllAnthemBorders();
		end
	end

	self.BuffWindow:UpdateBackground(anthemEffectPlace - 1);

	if (anthemEffectPlace <= 6 and isAnthemFixed == false) then
		for i = anthemEffectPlace, 6, 1 do
			self.BuffWindow:HideAnthem(i);
		end
	end
	--[[
	local party = LocalPlayer:GetParty();
	if (party ~= nil and party:GetMemberCount() > 1) then
		Turbine.Shell.WriteLine("Count: " .. party:GetMemberCount());
		local partyPlayer = party:GetMember(1);
		Turbine.Shell.WriteLine("Name: " .. partyPlayer:GetName());
		Turbine.Shell.WriteLine("Morale: " .. partyPlayer:GetMorale());
		--Turbine.Shell.WriteLine("Effects: " .. partyPlayer:GetEffects());
		
		--local partyEffectList = partyPlayer.EffectList;
		--DebugTable(party);		
		local partyEffectlist = partyPlayer:GetEffects();
		if (partyEffectlist ~= nil) then
			Turbine.Shell.WriteLine("Effects: " .. partyEffectlist:GetCount());
			--DebugTable(EffectList);
		end
		--local partyEffectCount = partyEffectList:GetCount();
		--Turbine.Shell.WriteLine("PartyEffects: " .. partyEffectCount);
	end
	]]--
end

--[[
function DebugTable(class)
	if type(class) ~= "table" then
	 Turbine.Shell.WriteLine( tostring(class) .. ": (" .. type(class) .. ")");
	 return;
	end
  local sortAtable = {}
  for key,v in pairs(class) do table.insert(sortAtable, {key,v}) end
  table.sort(sortAtable,compareItems);
	Turbine.Shell.WriteLine( tostring(class) .. ": (" .. type(class) .. ") " .. " : # " .. #sortAtable);
	cltb = getmetatable(class);
	for n,var in pairs(sortAtable) do
	 k = var[1]; v = var[2];
	 Turbine.Shell.WriteLine( "(" .. k .. "): (" .. type(v) .. ") " .. tostring(v) .. " ;" );
	 if k == "__implementation" then
	  u = getmetatable(v);	v3 = v;
	  if u then
	   for k2, v2 in pairs( u ) do
	    Turbine.Shell.WriteLine( "C: (" .. k2 .. "): " .. tostring(v2) );
	    if k2 == "__index" then v3 = v2; end
	   end
	   if v3 == v then
	    if u.__index then v3 = u.__index;
	    elseif cltb and type(v) == "userdata" then
	     if cltb.__index then v3 = cltb.__index;
	     else v3 = cltb; end
	    end
	   end
	  end
	  u = v3;
	 elseif k == "__index" and not cltb and getmetatable(v) then u = getmetatable(v).__index;
	 else u = v; end
--	 Turbine.Shell.WriteLine( "(" .. k .. "): (" .. type(u) .. ") " .. tostring(u) .. " ;" );
	 if type(u) == "table" then  -- can call DebugTable(u) recursively instead, but it's not advisable
	  table.sort(u);
	  for k2, v2 in pairs( u ) do
	    Turbine.Shell.WriteLine( "(" .. k2 .. "): " .. tostring(v2) );
	  end
	 elseif u ~= v then
	  Turbine.Shell.WriteLine( tostring(u) );
	 end
	end
	if cltb then
	 for k2, v2 in pairs( cltb ) do
	  Turbine.Shell.WriteLine( "  (" .. k2 .. "): " .. tostring(v2) );
	 end
	end
end
]]--

-- Returns true, if the effect is a valid ballad effect.
-- effectName - string
function MinstrelBuffMain:IsValidBalladEffect(effectName)
	return ValidBalladEffects[effectName] ~= nil;
end

-- Returns true, if the effect is a valid anthem effect.
-- effectName - string
function MinstrelBuffMain:IsValidAnthemEffect(effectName)
	return ValidAnthemEffects[effectName] ~= nil;
end

-- Returns the position of the effect.
-- effectName - string
function MinstrelBuffMain:GetAnthemEffectPosition(effectName)
	return ValidAnthemEffects[effectName];
end

function MinstrelBuffMain:IsValidWarSpeechEffect(effectName)
	return ValidWarSpeechSkillNames[effectName];
end

-- Melodic Interlude has been seen after the following skills:
--  Call to Fate, Dissonant Piercing Cry, Call of Oromë,
--  Cry of the Valar, Cry of the Wizards, Call of the Second Age
function MinstrelBuffMain:IsValidMelodicInterludeEffect(effectName)
	return ValidMelodicInterludeNames[effectName];
end

-- Returns true, if the skill is a minstrel coda.
-- skill - Turbine.Gameplay.ActiveSkill
function MinstrelBuffMain:IsValidCodaSkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	return ValidCodaSkillNames[skillName] ~= nil;
end

function MinstrelBuffMain:IsValidCrySkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	return ValidCrySkillNames[skillName] ~= nil;
end

function MinstrelBuffMain:IsValidMajorBalladSkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	return ValidMajorBalladSkillNames[skillName] ~= nil;
end

function MinstrelBuffMain:IsValidYellowTraitSkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	return ValidYellowTraitSkillNames[skillName] ~= nil;
end

-- Gets the count of the buffs.
function MinstrelBuffMain:GetBuffCount()
	local count = 0;
	count = EffectList:GetCount();
	return count;
end

-- Returns true, if the player is currently on a war steed.
function MinstrelBuffMain:IsPlayerOnWarsteed()
    local mount = LocalPlayer:GetMount();
    return mount ~= nil and mount:IsA(Turbine.Gameplay.CombatMount);
end

-- Occurs, when the minstrel stance was changed.
function MinstrelBuffMain:MinstrelStanceChanged()
	self.BuffWindow:SetAnthemBackgroundFromStance();
	--Turbine.Shell.WriteLine(self.classAttributes:GetStance());
end

-- Occurs, when a coda skill was used.
function MinstrelBuffMain:CodaSkillResetTimeChanged()
	self.codaWasPlayed = true; -- only mark coda as played, because the resetting event for the anthems and ballads can occur later.
	--self.BuffWindow:HideAllAnthemBorders();
end

-- Occurs, when a cry skill was used.
function MinstrelBuffMain:CrySkillResetTimeChanged()
	-- Tell the window about the event:
	self.BuffWindow:CryOfTheChorus();
end

-- Occurs, when a major ballad skill was used.
function MinstrelBuffMain:MajorBalladResetTimeChanged()
	-- Tell the window about the event:
	self.BuffWindow:MajorBallad();
end

-- Refreshes the visibility of the main window (i.e. when out of combat, no effects displayed).
function MinstrelBuffMain:RefreshVisibility()
	self.BuffWindow:SetVisibility();
end

-- Refreshes the window when the player is on a war steed or gets off a war steed.
function MinstrelBuffMain:RefreshMounted()
	self.settings:SetIsOnWarsteed(self:IsPlayerOnWarsteed());
end

mbMain = MinstrelBuffMain();
--mbCommand = MinstrelBuffs.MinstrelBuffWindow.MinstrelBuffCommand(mbMain);
MinstrelBuffCommand.MinstrelBuffMain = mbMain;
Turbine.Shell.AddCommand( "minstrelbuff", MinstrelBuffCommand);
Turbine.Shell.AddCommand( "mb", MinstrelBuffCommand);

mbMain:ShowWindow();
MinstrelBuffCommand:ShowHelpText();
